Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Qcow_cstructsSourceA subset of the Cstruct signature with type t = Cstruct.t list
This should be replaced with another parser, perhaps angstrom?
Returns a contiguous Cstruct.t, which may or may not involve a copy.