package qcow

  1. Overview
  2. Docs

Module Qcow_paddedSource

Sourcemodule Make (B : Qcow_s.RESIZABLE_BLOCK) : sig ... end

A block device which is padded with virtual zeroes so that reads beyond the current end don't fail.