package qcow

  1. Overview
  2. Docs

Module Qcow_rwlockSource

Sourcetype t

A lock which permits multiple concurrent threads to acquire it for reading but demands exclusivity for writing

Sourceval sexp_of_t : t -> Sexplib0.Sexp.t
Sourcetype ts = t list
Sourceval sexp_of_ts : ts -> Sexplib0.Sexp.t
Sourceval make : (unit -> string) -> t

make describe_fn creates a new lock, where describe_fn () returns a human-readable description string suitable for debug output.

Sourcetype lock

A value which represents holding a lock

Sourceval unlock : lock -> unit

unlock locked releases the lock associated with locked

Sourcemodule Client : sig ... end
Sourcemodule Read : sig ... end
Sourcemodule Write : sig ... end
Sourcemodule Debug : sig ... end