You can search for identifiers within the package.
in-package search v0.2.0
qcow
Qcow_padded
module Make (B : Qcow_s.RESIZABLE_BLOCK) : sig ... end
A block device which is padded with virtual zeroes so that reads beyond the current end don't fail.